import ngsolve as ngs
import netgen.csg as csg
import scipy.sparse as sparse
import numpy as np

# Construction of geometry and mesh
left  = csg.Plane(csg.Pnt(0,0,0), csg.Vec(-1,0,0) )
right = csg.Plane(csg.Pnt(1,1,1), csg.Vec( 1,0,0) )
front = csg.Plane(csg.Pnt(0,0,0), csg.Vec(0,-1,0) )
back  = csg.Plane(csg.Pnt(1,1,1), csg.Vec(0, 1,0) )
bot   = csg.Plane(csg.Pnt(0,0,0), csg.Vec(0,0,-1) )
top   = csg.Plane(csg.Pnt(1,1,1), csg.Vec(0,0, 1) )
left.bc('left')

cube = left * right * front * back * bot * top
geo = csg.CSGeometry()
geo.Add(cube)

mesh = ngs.Mesh(geo.GenerateMesh(maxh=0.1))

# Definition of function space
V = ngs.H1(mesh, order=1, dirichlet='left', dim=mesh.dim)
# V = ngs.VectorH1(mesh, order=2, dirichlet='left')

# Definition of bilinear forms
mu = 1  # all constants set to zero
lmbda = 1
rho = 1

def epsilon(u):
    return 0.5 * (ngs.Grad(u) + ngs.Grad(u).trans)

def sigma(u):
    return 2 * mu * epsilon(u) + lmbda * ngs.Trace(epsilon(u)) * ngs.Id(3)

u, v = V.TnT()
k = ngs.BilinearForm(V, symmetric=True)
k += ngs.InnerProduct(sigma(u), epsilon(v)) * ngs.dx
k.Assemble()

m = ngs.BilinearForm(V, symmetric=True)
m += rho * ngs.InnerProduct(u, v) * ngs.dx
m.Assemble()

# Conversion of matrices to sparse scipy matrices

if V.dim > 1:
    vals, cols, rows = k.mat.CSR()
    def ConvertB2Scal(BS, avals, acols, arows):
        BS = np.uint64(BS)
        BSs = np.uint64(BS**2)
        print(type(BS), type(BSs), type(avals[0]), type(acols[0]), type(arows[0]))
        vals = avals.NumPy()
        cols = acols.NumPy()
        rows = arows.NumPy()
        H = np.uint64(len(arows)-1)
        h = np.uint64(BS*H)
        svals = np.empty(len(vals))
        scols = np.empty(np.uint64(len(cols))*BSs, int)
        srows = np.empty(h+np.uint64(1), int)
        crow, ccol = np.uint64(0), np.uint64(0)
        one = np.uint64(1)
        for k in (np.uint64(x) for x in range(H)):
            bcols = cols[rows[k]:rows[k+one]]
            bvals = vals[BSs*rows[k]:BSs*rows[k+one]]
            for lk in (np.uint64(x) for x in range(BS)): # row BS*k+lk
                srows[crow] = int(ccol)
                for j in (np.uint64(x) for x in range(len(bcols))):
                    for lj in (np.uint64(x) for x in range(BS)):
                        scols[ccol] = int(BS * np.uint64(bcols[j]) + lj)
                        svals[ccol] = bvals[BSs*np.uint64(j) + BS*lj + lj]
                        ccol = ccol + one
                crow = crow + one
        srows[crow] = int(ccol)
        return svals, scols, srows
    svals, scols, srows = ConvertB2Scal(V.dim, vals, cols, rows)
else:
    svals, scols, srows = k.mat.CSR()

k_sp = sparse.csr_matrix((svals,scols,srows), shape=[V.ndof*V.dim, V.ndof*V.dim])

import matplotlib.pylab
matplotlib.pylab.spy(k_sp)
matplotlib.pylab.show()

# scalar cols, vals, rows

# Output
# <class 'ngsolve.bla.FlatVectorD'>
# <class 'pyngcore.FlatArray_I_S'>
# <class 'pyngcore.FlatArray_S_S'>
